@@ -48,7 +48,7 @@ class MorphValue(Keyword):
               '2', 'Dist', 'Bantu5', 'Prog', 'Pqp', 'Vnoun', 'Bantu4', 'Nfh', 'Emp', 'In', 'Pos', 'Des', 'Inan', 'Ord', 'Del', 'Add', 'Def', 'Ine',
               'Bantu10', 'Act', 'Bantu19', 'Ill', 'Imp', 'Art', 'Inv', 'Voc', 'Neut', 'Ind', 'Bantu14', 'Mid', 'Dem', 'Par', 'Past', 'Card', 'Fut',
               'Prs', 'Grpl', 'Lat', 'Antip', 'Perf', 'Yes', 'Bantu7', 'Tem', 'Ela', 'Spec', 'Nec', 'Tri', 'Ins', 'Cnd', 'Erg', 'Bantu9', 'Bantu12',
-              'Nhum', 'Bantu6', 'Sub', "Short", "Inen", "Ja", "Lainen", "Llinen", "Vs", "U", "Tar", "Digit"] + lang_spec_vals
+              'Nhum', 'Bantu6', 'Sub', "Short", "Inen", "Ja", "Lainen", "Llinen", "Vs", "U", "Tar", "Digit", "Combi", "Roman", "Word"] + lang_spec_vals
     grammar = Enum(*(K(v) for v in VALUES))
 
 
@@ -169,22 +169,48 @@ class Guard(Namespace):
     grammar = attr('bclauses', maybe_some(csl(BinaryClause))), optional(optional(","), attr('uclauses', maybe_some(csl(UnaryClause)))),\
         optional(blank), "->", optional(blank), attr('templates', restline)
 
+    def __str__(self):
+        if self.bclauses and self.uclauses:
+            return "{}, {} -> {}".format(", ".join(map(str, self.bclauses)), ", ".join(map(str, self.uclauses)), self.templates)
+        elif self.bclauses:
+            return "{} -> {}".format(", ".join(map(str, self.bclauses)), self.templates)
+        else:
+            return "{} -> {}".format(", ".join(map(str, self.uclauses)), self.templates)
+
+
 class SubtreePlaceholder:
     grammar = '<', name(), optional(attr('tag', DepTagChain)), optional(attr('neg', NegativeDepRel)), '>'
 
     def __str__(self):
-        return "<" + self.name + "." + ".".join([str(t) for t in list(self.tag)]) + (" - " + str(self.neg) if hasattr(self, 'neg') and len(self.neg) > 0 else "") + ">"
+        return "<{}.{}{}>".format(
+            self.name,
+            ".".join([str(t) for t in list(self.tag)]),
+            (" - {}".format(str(self.neg)) if hasattr(self, 'neg') and len(self.neg) > 0 else "")
+        )
 
 
 class NodePlaceholder:
     grammar = '[', name(), optional(attr('tag', DepTagChain)), optional(".", attr('prop', Prop)), ']'
 
     def __str__(self):
-        return "[" + self.name + "." + ".".join([str(t) for t in list(self.tag)]) + "]"
+        return "[{}.{}]".format(self.name, ".".join([str(t) for t in list(self.tag)])) if self.tag else "[{}]".format(self.name)
+
 
 class Template(List):
     grammar = some([question_word, inter_word, NodePlaceholder, SubtreePlaceholder]), optional(blank), "=>",\
-        optional(blank), attr('answer', some([inter_word, NodePlaceholder, SubtreePlaceholder]))
+        optional(blank), attr('answer', optional(some([inter_word, NodePlaceholder, SubtreePlaceholder])))
+
+    def __str__(self):
+        # NOTE: this will give you a non-normalized template (without IDs and possibly with repeating negatives)
+        return "{} => {}".format(" ".join(map(str, self)), " ".join(map(str, self.answer)))
+            
+    def to_string(self, only_answer=False, only_question=False):
+        if only_question:
+            return " ".join(map(str, self))
+        elif only_answer:
+            return " ".join(map(str, self.answer))
+        else:
+            return str(self)
 
 
 def filter_by_pos(node, value, f=None, check_subtrees=False, negate=False):
@@ -482,7 +508,6 @@ def placeholder2lambda(node, pholder):
             return fixed_cand
         else:
             # many of these occur for some reason!
-            print("HERE2", pholder)
             return None
     elif pholder.name == 'w':
         el = TemplateInstance(node)
@@ -574,5 +599,9 @@ def load_template_examples(files):
     return examples
 
 
-if __name__ == '__main__':
-    load_templates(['tmpl/en/templates/gen_templates.txt'])
\ No newline at end of file
+def parse_template(template):
+    return parse(template.strip(), Template)
+
+
+def parse_guard(guard):
+    return parse(guard.strip(), Guard)
\ No newline at end of file
